#!/bin/bash
#koi8-r
set +x

IPT_LEGACY=$(which iptables-legacy)
IPT_NFT=$(which iptables-nft)

[ "$2" = "-l" ] && IPT=${IPT_LEGACY} || IPT=${IPT_NFT}

[ -x "${IPT}" ] || {
	[ "$2" = "-l" ] && IPT=${IPT_NFT} || IPT=${IPT_LEGACY}
	[ -x "${IPT}" ] || {
		echo "ОШИБКА: Ни iptables-nft ни iptables-legacy не существуют, либо не исполняемые файлы."
		exit 1
	}
}

trap "term_exit" 2 3

clear_tables()
{
	local ipt=${1}

	${ipt} -F
	${ipt} -X
}

clean_tables()
{
	[ -x "${IPT_LEGACY}" ] && clear_tables ${IPT_LEGACY}
	[ -x "${IPT_NFT}" ] && clear_tables ${IPT_NFT}
}

term_exit()
{
	echo "Сброс конфигурации iptables"
	clean_tables
	exit 1
}

set_default_rules()
{
	$IPT -A INPUT  -j ACCEPT
	$IPT -A FORWARD -j ACCEPT
	$IPT -A OUTPUT -j ACCEPT
}

check_ping()
{
	#
	# ping ждёт 10 секунд и завершается.
	# Выставляем 12 сек на всякий случай.
	#
	ret=$(ping -c 1 -W 12 127.0.0.1 2>/dev/null)

	echo "${ret}" | grep -q "100% packet loss"
	return $?
}

test1()
{
	echo "ТЕСТ1. Проверка возможности блокировки исходящего сетевого трафика"

	clean_tables

	# Проверяем, что пинг проходит перед тестом
	check_ping && {
		echo -e "ICMP пакет отброшен до начала основного теста\nFAIL"
		clean_tables
		exit 1 
	}

	# Блокируем
	$IPT -A OUTPUT -j DROP -p icmp --icmp-type echo-request
	$IPT -A FORWARD -j ACCEPT
	$IPT -A INPUT -j ACCEPT

	if check_ping; then 
		echo -e "Исходящий трафик заблокирован\nPASS"
		clean_tables
		exit 0 
	else
		echo -e "Исходящий трафик открыт\nFAIL"
		clean_tables
		exit 1 
	fi	
}

test2()
{	
	echo "ТЕСТ2. Проверка возможности фильтрации сетевого трафика по IP-адресам"

	clean_tables
	
	# Проверяем, что пинг проходит перед тестом
	check_ping && {
		echo -e "ICMP пакет отброшен до начала основного теста\nFAIL"
		clean_tables
		exit 1 
	}

	# Блокируем
	$IPT -A OUTPUT -d 127.0.0.1 -j DROP
	$IPT -A FORWARD -j ACCEPT
	$IPT -A INPUT -j ACCEPT

	if check_ping; then 
		echo -e "Исходящий трафик к адресу 127.0.0.1 заблокирован\nPASS"
		clean_tables
		exit 0
	else
		echo -e "Исходящий трафик к адресу 127.0.0.1 открыт\nFAIL"
		clean_tables
		exit 1
	fi
}

test3()
{
	echo "ТЕСТ3. Проверка возможности фильтрации сетевого трафика по номеру сетевого интерфейса"

	clean_tables
	
	# Проверяем, что пинг проходит перед тестом
	check_ping && {
		echo -e "ICMP пакет отброшен до начала основного теста\nFAIL"
		clean_tables
		exit 1 
	}

	# Блокируем
	$IPT -A OUTPUT -o lo -j DROP
	$IPT -A FORWARD -j ACCEPT
	$IPT -A INPUT -j ACCEPT

	if check_ping; then 
		echo -e "Исходящий трафик через интерфейс loopback заблокирован\nPASS"
		clean_tables
		exit 0
	else
		echo -e "Исходящий трафик через интерфейс loopback открыт\nFAIL"
		clean_tables
		exit 1
	fi
}

test4()
{	
	echo "ТЕСТ4. Проверка возможности фильтрации сетевого трафика по сетевому протоколу и номеру порта"

	TTCP=${PWD}/bin/ttcp

	RET_FILE=${PWD}/RET_FILE
	rm -f ${RET_FILE}

	clean_tables

	$IPT -A OUTPUT -p udp --dport 5001 -j DROP
	$IPT -A FORWARD -j ACCEPT
	$IPT -A INPUT -j ACCEPT
	
	$TTCP -t -u 127.0.0.1 2>${RET_FILE}&
	sleep 5
	STATUS=`tail -n 5 ${RET_FILE}|grep "permitted"`

	if [ ! -z "${STATUS}" ];then
		echo -e "UDP-трафик заблокирован\nPASS"
		clean_tables
		NUM_PROC_STRING=`ps -ef | grep "ttcp" | awk '{print $2}'`
		kill -9 ${NUM_PROC_STRING} 2>/dev/null
		exit 0 
	else
		echo -e "UDP-трафик открыт\nFAIL"
		clean_tables
		NUM_PROC_STRING=`ps -ef | grep "ttcp" | awk '{print $2}'`
		kill -9 ${NUM_PROC_STRING} 2>/dev/null
		exit 1
	fi
}

start_tests()
{
	clean_tables
	set_default_rules
	test1
	test2
	test3
	test4
}

case "$1" in
start)	echo "Starting firewall_tc: ${IPT}"
	start_tests
	;;
stop) echo "Stopping firewall_tc: ${IPT}"
	clean_tables
	;;
test1) echo "Starting test1: ${IPT}"
	test1
	;;
test2) echo "Starting test2: ${IPT}"
	test2
	;;
test3) echo "Starting test3: ${IPT}"
	test3
	;;
test4) echo "Starting test4: ${IPT}"
	test4
	;;
esac

clean_tables

exit 0
	
