#!/usr/bin/python3

import os
import re
import sys
import time
import shutil
import atexit
import socket
import argparse
import subprocess
import multiprocessing
try:
    import PyPDF2
    PyPDF2_PRESENT = True
except ImportError:
    PyPDF2_PRESENT = False

TEST_NAME = 'marker_tc'

CUPS_FILES_CONF = '/etc/cups/cups-files.conf'
CUPS_PRINTERS_CONF = '/etc/cups/printers.conf'

TEST_PRINTER_NAME = 'marker_tc'
TEST_PRINTER_PATH = '/tmp/' + TEST_PRINTER_NAME + '.printer'

TEST_MAC_LEVEL = 2

TEST_JOB_NAME = TEST_PRINTER_NAME + '.txt'
TEST_PRINT_FILE = '/tmp/' + TEST_JOB_NAME

IS_VERBOSE = False

HOSTNAME = 'localhost'

TEST_RESULT = False

main_page_variables = {
    'mac-inv-num':'234-77-12'
}

last_page_variables = {
    'mac-owner-phone':'22-22',
    'mac-workplace-id':'Room_511',
    'mac-distribution':'In_archive'
}

class Logger:
    @staticmethod
    def _print(string):
        if IS_VERBOSE is True:
            print(time.strftime('%b %d %X') + ' ' + HOSTNAME + ' ' + TEST_NAME + ': ' + string)

    @staticmethod
    def info(string):
        Logger._print('[Info] ' + string)

    @staticmethod
    def error(string):
        Logger._print('[Error] ' + string)


def main_exit_trap():
    Logger.info('Тест завершен')

    if TEST_RESULT is True:
        Logger.info('ПРОВЕРКА МАРКИРОВКИ ПРОШЛА УСПЕШНО')
        print('PASS')
    else:
        Logger.info('ПРОВЕРКА МАРКИРОВКИ ЗАВЕРШИЛАСЬ ОШИБКОЙ')
        print('FAIL')

    shell_run('lprm -P ' + TEST_PRINTER_NAME)
    shell_run('lpoptions -x ' + TEST_PRINTER_NAME)
    shell_run('lpadmin -x ' + TEST_PRINTER_NAME)
    if os.path.exists(TEST_PRINTER_PATH) is True:
        os.remove(TEST_PRINTER_PATH)
    shell_run('/etc/init.d/cups restart')


def is_running(proc_name):
    i = 0
    while i < 5:
        i += 1
        time.sleep(3)
    result = _is_running(proc_name)
    return result


def _is_running(proc_name):
    try:
        ret = subprocess.check_call('ps -ef | grep -v grep | grep cupsd', shell=True,
                                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        ret = 1

    if ret == 0:
        result = True
    else:
        result = False
    return result


def shell_run(cmd):
    subprocess.call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)


def shell_run_with_result(cmd):
    try:
        ret = subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True)
    except subprocess.CalledProcessError:
        ret = 1

    if ret == 0:
        result = True
    else:
        result = False
    return result


def shell_run_with_stdout(cmd):
    try:
        res = subprocess.check_output(cmd, stderr=subprocess.DEVNULL, shell=True, universal_newlines=True).strip()
    except subprocess.CalledProcessError:
        res = ''
    return res

def marker_type():
    type = 'Binary'
    Logger.info('Определение типа маркера')
    res = shell_run_with_stdout('elmac-marker-configure get-current-type')
    if res == 'Python':
        if PyPDF2_PRESENT == False:
            Logger.info('Модуль PyPDF2 не найден. Маркер Python неработоспособен')
            sys.exit(1)
        type = res
    Logger.info('Используется ' + type + ' маркер')
    return type


def main():
    atexit.register(main_exit_trap)

    args_parser = argparse.ArgumentParser()
    args_parser.add_argument('-v', '--verbose', dest='is_verbose', action='store_true', help='Enable verbose mode')
    args = args_parser.parse_args()
    global IS_VERBOSE
    IS_VERBOSE = args.is_verbose

    global HOSTNAME
    HOSTNAME = socket.gethostname()

    Logger.info('Запуск теста')
    marker_type()

    if os.getuid() != 0:
        Logger.error('Требуется запускать тест от пользователя root!')
        sys.exit(1)

    if os.path.exists(CUPS_FILES_CONF) is False:
        if os.path.exists(CUPS_FILES_CONF + '.default') is False:
            Logger.error('Файл \"' + CUPS_FILES_CONF + '.default\" не найден!')
            sys.exit(1)
        shutil.copyfile(CUPS_FILES_CONF + '.default', CUPS_FILES_CONF)

    if os.path.exists(CUPS_FILES_CONF) is False:
        Logger.error('Файл \"' + CUPS_FILES_CONF + '\" не найден!')
        sys.exit(1)

    with open(CUPS_FILES_CONF, 'r') as file:
        file_lines = file.readlines()

    is_changed = False
    file_device_base_regex = re.compile('FileDevice')
    file_device_true_regex = re.compile('^FileDevice Yes$')
    system_group_regex = re.compile('^SystemGroup(.*)\s$')
    i = 0
    for line in file_lines:
        if re.search(file_device_base_regex, line):
            if re.search(file_device_true_regex, line) is None:
                file_lines.pop(i)
                file_lines.insert(i, 'FileDevice Yes\n')
                is_changed = True
        if re.search(system_group_regex, line):
            file_lines.pop(i)
            file_lines.insert(i, '#' + line)
            is_changed = True
        i += 1

    if is_changed is True:
        Logger.info('Добавление возможности печати в файл (FileDevice Yes)')
        data = ''.join(file_lines)
        with open(CUPS_FILES_CONF, 'w') as file:
            file.write(data)

    Logger.info('Перезапуск cupsd')
    shell_run('/etc/init.d/cups restart')

    if is_running('cupsd') is False:
        Logger.error('Процесс cupsd не запущен!')
        sys.exit(1)

    shell_run('lpadmin -p ' + TEST_PRINTER_NAME +
              ' -v file://' + TEST_PRINTER_PATH + ' -m drv:///sample.drv/generic.ppd -E')
    Logger.info('Создание принтера ' + TEST_PRINTER_NAME)

    shell_run('lpadmin -p ' + TEST_PRINTER_NAME + ' -o printer-op-policy=elmac-policy -E')
    shell_run('lpoptions -p ' + TEST_PRINTER_NAME + ' -o media=A4')

    is_printer_added = False
    printer_regex = re.compile('^<Printer ' + TEST_PRINTER_NAME + '>$')
    i = 0
    while i < 60:
        i += 1
        time.sleep(2)
        if os.path.exists(CUPS_PRINTERS_CONF) is False:
            continue
        with open(CUPS_PRINTERS_CONF, 'r') as file:
            lines = file.readlines()
            for line in lines:
                if re.search(printer_regex, line):
                    is_printer_added = True
                    break
        if is_printer_added is True:
            break

    if is_printer_added is False:
        Logger.error('Ошибка добавления принтера!')
        sys.exit(1)

    if os.path.exists(CUPS_PRINTERS_CONF) is False:
        Logger.error('Файл \"' + CUPS_PRINTERS_CONF + '\" не найден!')
        sys.exit(1)

    with open(CUPS_PRINTERS_CONF, 'r') as file:
        file_lines = file.readlines()

    is_section = False
    begin_section = re.compile('^<Printer ' + TEST_PRINTER_NAME + '>$')
    end_section = re.compile('^</Printer>$')
    printer_mac_max_label = re.compile('^MacLabelMax(.*)\s$')
    printer_mac_policy = re.compile('^MacPolicy(.*)\s$')
    i = 0
    for line in file_lines:
        if re.search(begin_section, line):
            is_section = True
        if re.search(end_section, line):
            is_section = False
        if is_section is True:
            if re.search(printer_mac_max_label, line):
                file_lines.pop(i)
                file_lines.insert(i, 'MacLabelMax ' + str(TEST_MAC_LEVEL) + ':0x0\n')
            if re.search(printer_mac_policy, line):
                file_lines.pop(i)
                file_lines.insert(i, 'MacPolicy wr-up\n')
        i += 1

    Logger.info('Остановка cupsd!')
    shell_run('/etc/init.d/cups stop')
    if is_running('cupsd') is True:
        Logger.error('Не удалось завершить процесс cupsd!')
        sys.exit(1)

    data = ''.join(file_lines)
    with open(CUPS_PRINTERS_CONF, 'w') as file:
        file.write(data)
        Logger.info('Установка принтеру ' + TEST_PRINTER_NAME + ' мандатного уровня ' + str(TEST_MAC_LEVEL))

    Logger.info('Запуск cupsd')
    shell_run('/etc/init.d/cups restart')

    if is_running('cupsd') is False:
        Logger.error('Процесс cupsd не стартовал!')
        sys.exit(1)

    proc = multiprocessing.Process(target=do_printing, args=())
    proc.start()
    proc.join()

    global TEST_RESULT
    if proc.exitcode == 0:
        TEST_RESULT = True
        sys.exit(0)
    else:
        TEST_RESULT = False
        sys.exit(1)


def do_printing():
    from ctypes import cdll, Structure, c_ubyte, c_longlong, pointer
    lib = cdll.LoadLibrary('libelmac.so.0')

    # struct from /usr/include/linux/elmac.h
    class elmac_label_t(Structure):
        _fields_ = [('level', c_ubyte),
                    ('category', c_longlong)]

    c_level = c_ubyte(TEST_MAC_LEVEL)
    content_set = elmac_label_t(c_level, 0)
    content_set_p = pointer(content_set)
    res = lib.elmac_set_label_for_self(content_set_p)
    if res != 0:
        Logger.error('Ошибка установки мандатного уровня!')
        sys.exit(1)
    else:
        Logger.info('Установка процессу мандатного уровня ' + str(TEST_MAC_LEVEL))

    with open(TEST_PRINT_FILE, 'w') as file:
        file.write('Test marker_tc file. Level ' + str(TEST_MAC_LEVEL) + '\n')

    if os.path.exists(TEST_PRINTER_PATH) is True:
        os.remove(TEST_PRINTER_PATH)

    Logger.info('Запуск печати тестового файла')
    res = shell_run_with_stdout('lp -d ' + TEST_PRINTER_NAME + ' ' + TEST_PRINT_FILE)

    if len(res) == 0:
        Logger.error('Не удалось напечатать тестовый документ!')
        sys.exit(1)

    res = str(res)
    res = res.split(TEST_PRINTER_NAME + '-')[-1]
    job_id = res.split(' ')[0]

    if len(job_id) == 0:
        Logger.error('Не удалось получить job id!')
        sys.exit(1)
    else:
        Logger.info('job id задания в очереди - ' + str(job_id))

    Logger.info('Маркировка задания ' + str(job_id))
    for key in main_page_variables:
        shell_run('lpattr -j ' + job_id + ' -s ' + key + '=\"\\"' +
                main_page_variables[key] + '\\"\"')
    for key in last_page_variables:
        shell_run('lpattr -j ' + job_id + ' -s ' + key + '=\"\\"' +
                last_page_variables[key] + '\\"\"')
    shell_run('lpattr -j ' + job_id + ' -m')

    content_set = elmac_label_t(0, 0)
    content_set_p = pointer(content_set)
    res = lib.elmac_set_label_for_self(content_set_p)
    if res != 0:
        Logger.error('Ошибка установки мандатного уровня!')
        sys.exit(1)
    else:
        Logger.info('Установка процессу мандатного уровня ' + str(0))

    main_job_id = None
    last_job_id = None
    main_job_regex = re.compile('^(.*)mac\(' + job_id + '\)(.*)' + TEST_JOB_NAME + '(.*)$')
    last_job_regex = re.compile('^(.*)mac\*\(' + job_id + '\)(.*)' + TEST_JOB_NAME + '(.*)$')
    i = 0
    while i < 60:
        i += 1
        time.sleep(1)
        res = shell_run_with_stdout('lpq -P ' + TEST_PRINTER_NAME)
        all_jobs = res.split('\n')
        for line in all_jobs:
            if re.search(main_job_regex, line):
                l = line.split()
                main_job_id = l[2]
            if re.search(last_job_regex, line):
                l = line.split()
                last_job_id = l[2]
            if main_job_id is not None and last_job_id is not None:
                break

    if main_job_id is None or last_job_id is None:
        Logger.error('Не удалось получить промаркированные job id!')
        sys.exit(1)
    else:
        Logger.info('Промаркированные задания ' + str(main_job_id) + ' ' + str(last_job_id))

    Logger.info('Печать основного задания в файл')
    shell_run('lp -i ' + str(main_job_id) + ' -H resume')
    time.sleep(5)

    main_file = '/tmp/' + TEST_JOB_NAME + '.main.ps'
    if os.path.exists(main_file) is True:
        os.remove(main_file)

    timeout = time.time() + 5
    while time.time() < timeout:
        if os.path.exists(TEST_PRINTER_PATH) is True:
            shutil.copyfile(TEST_PRINTER_PATH, main_file)
            os.remove(TEST_PRINTER_PATH)
            break

    Logger.info('Печать оборота задания в файл')
    shell_run('lp -i ' + str(last_job_id) + ' -H resume')
    time.sleep(5)

    last_file = '/tmp/' + TEST_JOB_NAME + '.last.ps'
    if os.path.exists(last_file) is True:
        os.remove(last_file)

    timeout = time.time() + 5
    while time.time() < timeout:
        if os.path.exists(TEST_PRINTER_PATH) is True:
            shutil.copyfile(TEST_PRINTER_PATH, last_file)
            os.remove(TEST_PRINTER_PATH)
            break

    if marker_type() == 'Python':
        check_marked_pdf(main_file)
        check_marked_pdf(last_file)
    else:
        check_marked_ps(main_file)
        check_marked_ps(last_file)


def check_marked_ps(in_file):
    if in_file == '/tmp/' + TEST_JOB_NAME + '.main.ps':
        Logger.info('Проверка маркировки основного задания')
        self_marker = re.compile('^/TopPos TopPos tPart sub bPart sub markerStringInterval sub def$')
    elif in_file == '/tmp/' + TEST_JOB_NAME + '.last.ps':
        Logger.info('Проверка маркировки оборота задания')

        # disable back side check cuz new cups marker create new ps format file
        Logger.info('Вручную проверьте корректность маркировки командой:')
        Logger.info('      \"evince ' + in_file + '\"')
        return

        self_marker = re.compile('^/BottomPos BottomPos tPart add bPart add markerStringInterval add def$')
    else:
        Logger.error('Указан некоректный файл для проверки маркировки!')
        sys.exit(1)

    is_self_marker = False
    is_base_marker1 = False
    base_marker1 = re.compile('^initclip newpath clippath pathbbox closepath$')
    is_base_marker2 = False
    base_marker2 = re.compile('^/pageT exch def /pageR exch def /pageB exch def /pageL exch def$')

    with open(in_file, 'r') as file:
        file_lines = file.readlines()
        for line in file_lines:
            if re.search(base_marker1, line):
                is_base_marker1 = True
            if re.search(base_marker2, line):
                is_base_marker2 = True
            if re.search(self_marker, line):
                is_self_marker = True

    if is_base_marker1 is False and is_base_marker2 is False and is_self_marker is False:
        Logger.error('Индивидуальный маркер - ' + str(is_self_marker))
        Logger.error('Базовый маркер 1 - ' + str(is_base_marker1))
        Logger.error('Базовый маркер 2 - ' + str(is_base_marker2))
        Logger.error('Маркировка в файле ' + in_file + ' не найдена!')
        sys.exit(1)
    else:
        Logger.info('Вручную проверьте корректность маркировки командой:')
        Logger.info('      \"evince ' + in_file + '\"')


def marker_is_present_in_pdf(file_pdf, marker):
    pdfFileObj = open(file_pdf, 'rb')
    pdfReader = PyPDF2.PdfFileReader(pdfFileObj, strict=False)
    search_word = marker
    search_word_count = 0

    for pageNum in range(0, pdfReader.numPages):
        pageObj = pdfReader.getPage(pageNum)
        text = pageObj.extractText().encode('utf-8')
        #search_text = text.lower().split()
        search_text = text.split()
        for word in search_text:
            if search_word in word.decode("utf-8"):
                search_word_count += 1
    if search_word_count > 0:
        return True

    return False


def check_marked_pdf(in_file):
    file_ps = ''
    file_pdf = ''
    page_variables = {}

    if in_file == '/tmp/' + TEST_JOB_NAME + '.main.ps':
        Logger.info('Проверка маркировки основного задания')
        file_ps = '/tmp/' + TEST_JOB_NAME + '.main.ps'
        file_pdf = '/tmp/' + TEST_JOB_NAME + '.main.pdf'
        page_variables = main_page_variables
    elif in_file == '/tmp/' + TEST_JOB_NAME + '.last.ps':
        Logger.info('Проверка маркировки оборота задания')
        file_ps = '/tmp/' + TEST_JOB_NAME + '.last.ps'
        file_pdf = '/tmp/' + TEST_JOB_NAME + '.last.pdf'
        page_variables = last_page_variables
    else:
        Logger.error('Указан некоректный файл для проверки маркировки!')
        sys.exit(1)

    res = shell_run_with_result('ps2pdf13 ' + file_ps + ' ' + file_pdf)
    if res is False:
        Logger.error('Не удалось получить ' + file_pdf + '!')
        sys.exit(1)

    for key in page_variables:
        res = marker_is_present_in_pdf(file_pdf, page_variables[key])
        if res is False:
            Logger.error('Маркер ' + key + ':' + page_variables[key] + ' в файле '
                        + in_file + ' не найден!')
            sys.exit(1)
        Logger.info('Маркер ' + key + ':' + page_variables[key] + ' в файле '
                    + in_file + ' найден')


    Logger.info('Вручную проверьте корректность маркировки командой:')
    Logger.info('      \"evince ' + in_file + '\"')



if __name__ == '__main__':
    main()

