#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <stdarg.h>
#include <libgen.h>
#include <locale.h>
#include <libintl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>

#define _(msgid)	gettext(msgid)

#define MIN_SIZE	0x1
#define MAX_SIZE	0x100000
#define DEF_SIZE	0x100000
#define MIN_COUNT	0
#define MAX_COUNT	1000000
#define DEF_COUNT	100

static void usage(char *name)
{
	printf(_("Usage: %s [-h, --help] [-v, --version] [-s, --size <hex byte>] "
		"[-c, --count <count iteration>] -m, --method\n"
		"options:\n"
		" -v, --version       output version information and exit\n"
		" -h, --help          output usage information and exit\n"
		" -d, --debug         debug information\n"
		" -s, --size          memory block size (max 0x10000. Not relevant COW.)\n"
		" -c, --count         count allocated (if 0, allocated all free memmory. Not relevant COW.)\n"
		" -m, --method        tested method (allocated: malloc, sbrk, mmap isolation: COW\n)"), name);
}

static const char short_options[] = "hvds:c:m:";

static struct option options[] = {
	{"help", no_argument, 0, 'h'},
	{"version", no_argument, 0, 'v'},
	{"debug", no_argument, 0, 'd'},
	{"size", required_argument, 0, 's'},
	{"count", required_argument, 0, 'c'},
	{"method", required_argument, 0, 'm'},
	{NULL, 0, 0, 0}
};


static int	size = DEF_SIZE;
static int	count = DEF_COUNT;
static int	debug;
static char	*progname;

static void error(int val, const char *fmt, ...)
{
	va_list	ap;

	fprintf(stderr, _("%s: ERROR: "), progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
	exit(val);
}

/*
 * Convert size mem.
 */
static void conv_size(char *val)
{
	int dummy;
	int rc;

	rc = sscanf(val, "%x", &dummy);
	if(rc != 1)
		error(-ERANGE, _("bad len value"));
	size = dummy;
	if ((size > MAX_SIZE) || (size < MIN_SIZE)) {
		char err[4096] = {0};
		snprintf(err, sizeof(err) - 1, _("bad size value, min = %d max = %d\n"),
			 MIN_SIZE, MAX_SIZE);
		error(-ERANGE, err);
	}
}

/*
 * Convert count.
 */
static void conv_count(char *val)
{
	int dummy;
	int rc;

	rc = sscanf(val, "%d", &dummy);
	if(rc != 1)
		error(-ERANGE, _("bad count value"));
	count = dummy;
	if ((count > MAX_COUNT) || (count < MIN_COUNT)) {
		char err[4096] = {0};
		snprintf(err, sizeof(err) - 1, _("bad count value, min = %d, max = %d\n"),
			 MIN_COUNT, MAX_COUNT);
		error(-ERANGE, err);
	}
}

/*
 * Cheking mmap
 */
int mmap_alloc()
{
	char* ptr_a;
	int i;
	int j;
	i = 0;

	printf("Test mmap.\n");
	printf("pid: %d size block: %x count iteration: %d\n",
	       getpid(), size, count);
	while (ptr_a = mmap(0, size, PROT_READ | PROT_WRITE,
	       MAP_PRIVATE | MAP_ANONYMOUS, -1, 0)) {
		if (debug)
			printf("DEBUG (mmap): pid: %d size block: %x addr: %p\n",
			       getpid(), size, ptr_a);
		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				printf("iteration #%d "
					"found pattern at %d [%c(%x)]\n",
					i, j, ptr_a[j], ptr_a[j]);
				exit(1);
			}
		}
		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				ptr_a[j] = 0xff;
			}
		}

		i++;
		if (i == count)
			return 0;
	}
	return -ENOMEM;
}

/*
 * Cheking sbrk
 */
int brk_alloc()
{
	char* ptr_a;
	int i;
	int j;

	i = 0;

	printf("Test sbrk.\n");
	printf("pid: %d size block: %x count iteration: %d\n",
	       getpid(), size, count);
	while (ptr_a = sbrk(size)) {
		if (debug)
			printf("DEBUG (sbrk): pid: %d size block: %x addr: %p\n",
			       getpid(), size, ptr_a);
		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				printf("iteration #%d "
					"found pattern at %d [%c(%x)]\n",
					i, j, ptr_a[j], ptr_a[j]);
				exit(1);
			}
		}

		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				ptr_a[j] = 0xff;
			}
		}

		i++;
		if (i == count)
			return 0;
        }
        return -ENOMEM;
}

/*
 * Cheking malloc
 */
int memalloc()
{
	char* ptr_a;
	int   i;
	int   j;

	printf("Test malloc function.\n");
	printf("pid: %d size block: %x count iteration: %d\n",
	       getpid(), size, count);
	
	i = 0;
	while (ptr_a = malloc(size)) {
		if (debug)
			printf("DEBUG (malloc): pid: %d size block: %x addr: %p\n",
			       getpid(), size, ptr_a);
		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				printf("iteration #%d "
				       "found pattern at %d [%c(%x)]\n",
				       i, j, ptr_a[j], ptr_a[j]);
				exit(1);
			}
		}

		for (j = 0; j < size; j++) {
			if (ptr_a[j] != 0x0) {
				ptr_a[j] = 0xff;
			}
		}

		i++;
		if (i == count)
			return 0;
	}

	return -ENOMEM;
}

#define sig_A	"COW proc A."
#define sig_B	"COW proc B."
int main(int argc, char **argv)
{
	size_t len;
	char *cow;
	int pid;
	int status;
	int c;
	int method = -1;
	int pfd[2];

	progname = basename(argv[0]);
	
	setlocale(LC_ALL, "");
	bindtextdomain(progname, "/usr/share/locale");
	textdomain(progname);


	for ( ; ; ) {
		c = getopt_long(argc, argv, short_options, options, NULL);

		if (c == -1)
			break;

		switch (c) {
		case 'h':
			usage(progname);
			exit(0);
		case 'v':
			printf("%s 0.1\n",
			       progname);
			exit(0);
		case 'd':
			debug = 1;
			break;
		case 's':
			conv_size(optarg);
			break;
		case 'c':
			conv_count(optarg);
			break;
		case 'm':
			if (!strcmp(optarg, "malloc"))
				method = 0;
			if (!strcmp(optarg, "sbrk"))
				method = 1;
			if (!strcmp(optarg, "mmap"))
				method = 2;
			if (!strcmp(optarg, "COW"))
				method = 3;
			break;
		default:
			usage(progname);
			exit(1);
		}
	}

	if ((method < 0) || (method > 3)) {
		error(-ERANGE, _("bad method name"));
	}
	switch (method) {
	case 0:
		if (memalloc() == -ENOMEM) {
			printf("pid: %d No free memory for allocation\n", getpid());
		}
		return 0;
		break;
	case 1:
		if (brk_alloc() == -ENOMEM) {
			printf("pid: %d No free memory for allocation\n", getpid());
		}
		return 0;
		break;
	case 2:
		if (mmap_alloc() == -ENOMEM) {
			printf("pid: %d No free memory for allocation\n", getpid());
		}
		return 0;
		break;
	}

	char bufp[10000];
	
	pipe(pfd);
	
	memset(bufp, 0, sizeof(bufp));
	snprintf(bufp, strlen(bufp) - 1, "Test COW.\n");
	
	len = strlen(sig_A);
	cow = (char *) malloc(len + 1);
	memcpy(cow, sig_A, len);

	snprintf(bufp + strlen(bufp), sizeof(bufp)- strlen(bufp) - 1,
		 "pid: %d cow addr: %p (\"%s\")\n",
		 getpid(), cow, cow);

	pid = fork();
	if (pid < 0)
		exit(1);
	if (pid == 0) {
		char bufc[1000];
		size_t len;

		close(pfd[0]);

		memset(bufc, 0, sizeof(bufc));
		snprintf(bufc, sizeof(bufc) - 1, "pid: %d cow addr: %p (\"%s\")\n", getpid(), cow, cow);

		len = strlen(sig_B);
		memcpy(cow, sig_B, len);

		snprintf(bufc + strlen(bufc), sizeof(bufc) - strlen(bufc) - 1,
			 "pid: %d cow addr: %p (\"%s\")\n", getpid(), cow, cow);

		len = strlen(bufc);
		write(pfd[1], &len, sizeof(len));
		write(pfd[1], (void *) bufc, strlen(bufc));
		exit(0);
	} else {
		int rc;
		size_t len;
		char *msg;

		close(pfd[1]);

		while (((rc = waitpid(pid, &status, 0)) == (pid_t)-1) &&
			(errno == EINTR));
		if (rc == (pid_t)-1) {
			exit(1);
		}
		if (!WIFEXITED(status) || WIFSIGNALED(status) > 0) {
			exit(1);
		}

		read(pfd[0], &len, sizeof(len));
		if (len > 0) {
			msg = (char *) malloc(len + 1);
			if (msg == NULL) {
				error(-ENOMEM, "Memmory alloc error");
			}
			memset(msg, 0, len + 1);
			while (read(pfd[0], msg, len + 1) != 0);;
			snprintf(bufp + strlen(bufp), sizeof(bufp)- strlen(bufp) - 1,
				 "%s\n", msg);
		}
		close(pfd[0]);

		snprintf(bufp + strlen(bufp), sizeof(bufp) - strlen(bufp) - 1,
			 "pid: %d cow addr: %p (\"%s\") after complite process B.\n",
			 getpid(), cow, cow);

		rc = 0;
		if (memcmp(cow, sig_A, strlen(sig_A))) {
			snprintf(bufp + strlen(bufp), sizeof(bufp) - strlen(bufp) - 1,
				 "FAIL pid: %d cow addr: %p (\"%s\" != \"%s\")\n",
			         getpid(), cow, cow, sig_A);
			rc = -1;
		}

		snprintf(bufp + strlen(bufp), sizeof(bufp) - strlen(bufp) - 1,
			"SUCCES pid: %d cow addr: %p (\"%s\" == \"%s\")\n",
			getpid(), cow, cow, sig_A);

		write(1, bufp, strlen(bufp));

		return rc;
	}
}
